From 58abc7effb90c0f884374028f807fd4af1998276 Mon Sep 17 00:00:00 2001
From: Sebastian Pop <spop@amazon.com>
Date: Thu, 29 Jul 2021 20:06:07 +0000
Subject: [PATCH] [arm64] port add_ps

         add_ps[  4x4]	20.65x 	 8.41     	 173.78
         add_ps[  8x8]	56.68x 	 12.71    	 720.24
 [i420]  add_ps[  4x4]	22.10x 	 9.22     	 203.88
 [i422]  add_ps[  4x8]	29.01x 	 15.05    	 436.49
         add_ps[16x16]	95.32x 	 39.01    	 3718.96
 [i420]  add_ps[  8x8]	65.08x 	 12.88    	 837.86
 [i422]  add_ps[ 8x16]	74.67x 	 23.75    	 1773.13
         add_ps[32x32]	119.22x 	 146.68   	 17488.19
 [i420]  add_ps[16x16]	102.74x 	 39.14    	 4020.90
 [i422]  add_ps[16x32]	112.34x 	 74.15    	 8330.32
         add_ps[64x64]	115.99x 	 596.99   	 69245.54
 [i420]  add_ps[32x32]	119.62x 	 146.92   	 17575.66
 [i422]  add_ps[32x64]	124.33x 	 291.17   	 36200.63
---
 source/common/aarch64/asm-primitives.cpp |  17 +++
 source/common/aarch64/pixel-util.S       | 183 +++++++++++++++++++++++
 2 files changed, 200 insertions(+)

diff --git a/source/common/aarch64/asm-primitives.cpp b/source/common/aarch64/asm-primitives.cpp
index 5e57f31..126bbaf 100644
--- a/source/common/aarch64/asm-primitives.cpp
+++ b/source/common/aarch64/asm-primitives.cpp
@@ -385,6 +385,23 @@ void setupAssemblyPrimitives(EncoderPrimitives &p, int cpuMask)
         p.chroma[X265_CSP_I422].cu[BLOCK_422_16x32].copy_sp = PFX(blockcopy_sp_16x32_neon);
         p.chroma[X265_CSP_I422].cu[BLOCK_422_32x64].copy_sp = PFX(blockcopy_sp_32x64_neon);
 
+        // pixel_add_ps
+        p.cu[BLOCK_4x4].add_ps[NONALIGNED]   = PFX(pixel_add_ps_4x4_neon);
+        p.cu[BLOCK_8x8].add_ps[NONALIGNED]   = PFX(pixel_add_ps_8x8_neon);
+        p.cu[BLOCK_16x16].add_ps[NONALIGNED] = PFX(pixel_add_ps_16x16_neon);
+        p.cu[BLOCK_32x32].add_ps[NONALIGNED] = PFX(pixel_add_ps_32x32_neon);
+        p.cu[BLOCK_64x64].add_ps[NONALIGNED] = PFX(pixel_add_ps_64x64_neon);
+
+        // chroma add_ps
+        p.chroma[X265_CSP_I420].cu[BLOCK_420_4x4].add_ps[NONALIGNED]   = PFX(pixel_add_ps_4x4_neon);
+        p.chroma[X265_CSP_I420].cu[BLOCK_420_8x8].add_ps[NONALIGNED]   = PFX(pixel_add_ps_8x8_neon);
+        p.chroma[X265_CSP_I420].cu[BLOCK_420_16x16].add_ps[NONALIGNED] = PFX(pixel_add_ps_16x16_neon);
+        p.chroma[X265_CSP_I420].cu[BLOCK_420_32x32].add_ps[NONALIGNED] = PFX(pixel_add_ps_32x32_neon);
+        p.chroma[X265_CSP_I422].cu[BLOCK_422_4x8].add_ps[NONALIGNED]   = PFX(pixel_add_ps_4x8_neon);
+        p.chroma[X265_CSP_I422].cu[BLOCK_422_8x16].add_ps[NONALIGNED]  = PFX(pixel_add_ps_8x16_neon);
+        p.chroma[X265_CSP_I422].cu[BLOCK_422_16x32].add_ps[NONALIGNED] = PFX(pixel_add_ps_16x32_neon);
+        p.chroma[X265_CSP_I422].cu[BLOCK_422_32x64].add_ps[NONALIGNED] = PFX(pixel_add_ps_32x64_neon);
+
         // cpy2Dto1D_shl
         p.cu[BLOCK_4x4].cpy2Dto1D_shl   = PFX(cpy2Dto1D_shl_4x4_neon);
         p.cu[BLOCK_8x8].cpy2Dto1D_shl   = PFX(cpy2Dto1D_shl_8x8_neon);
diff --git a/source/common/aarch64/pixel-util.S b/source/common/aarch64/pixel-util.S
index ea03ed3..f901d9d 100644
--- a/source/common/aarch64/pixel-util.S
+++ b/source/common/aarch64/pixel-util.S
@@ -362,6 +362,189 @@ function x265_pixel_sub_ps_32x64_neon
     ret
 endfunc
 
+// void x265_pixel_add_ps_neon(pixel* a, intptr_t dstride, const pixel* b0, const int16_t* b1, intptr_t sstride0, intptr_t sstride1);
+function x265_pixel_add_ps_4x4_neon
+    lsl             x5, x5, #1
+.rept 2
+    ld1             {v0.8b}, [x2], x4
+    ld1             {v1.8b}, [x2], x4
+    ld1             {v2.4h}, [x3], x5
+    ld1             {v3.4h}, [x3], x5
+    uxtl            v0.8h, v0.8b
+    uxtl            v1.8h, v1.8b
+    add             v4.8h, v0.8h, v2.8h
+    add             v5.8h, v1.8h, v3.8h
+    sqxtun          v4.8b, v4.8h
+    sqxtun          v5.8b, v5.8h
+    st1             {v4.s}[0], [x0], x1
+    st1             {v5.s}[0], [x0], x1
+.endr
+    ret
+endfunc
+
+function x265_pixel_add_ps_8x8_neon
+    lsl             x5, x5, #1
+.rept 4
+    ld1             {v0.8b}, [x2], x4
+    ld1             {v1.8b}, [x2], x4
+    ld1             {v2.8h}, [x3], x5
+    ld1             {v3.8h}, [x3], x5
+    uxtl            v0.8h, v0.8b
+    uxtl            v1.8h, v1.8b
+    add             v4.8h, v0.8h, v2.8h
+    add             v5.8h, v1.8h, v3.8h
+    sqxtun          v4.8b, v4.8h
+    sqxtun          v5.8b, v5.8h
+    st1             {v4.8b}, [x0], x1
+    st1             {v5.8b}, [x0], x1
+.endr
+    ret
+endfunc
+
+.macro pixel_add_ps_16xN_neon h i
+function x265_pixel_add_ps_16x\h\()_neon
+    lsl             x5, x5, #1
+    mov             w12, #\i
+.loop_add_ps_16x\h\():
+    sub             w12, w12, #1
+.rept 4
+    ld1             {v0.16b}, [x2], x4
+    ld1             {v1.16b}, [x2], x4
+    ld1             {v16.8h-v17.8h}, [x3], x5
+    ld1             {v18.8h-v19.8h}, [x3], x5
+    uxtl            v4.8h, v0.8b
+    uxtl2           v5.8h, v0.16b
+    uxtl            v6.8h, v1.8b
+    uxtl2           v7.8h, v1.16b
+    add             v24.8h, v4.8h, v16.8h
+    add             v25.8h, v5.8h, v17.8h
+    add             v26.8h, v6.8h, v18.8h
+    add             v27.8h, v7.8h, v19.8h
+    sqxtun          v4.8b, v24.8h
+    sqxtun2         v4.16b, v25.8h
+    sqxtun          v5.8b, v26.8h
+    sqxtun2         v5.16b, v27.8h
+    st1             {v4.16b}, [x0], x1
+    st1             {v5.16b}, [x0], x1
+.endr
+    cbnz            w12, .loop_add_ps_16x\h
+    ret
+endfunc
+.endm
+
+pixel_add_ps_16xN_neon 16 2
+pixel_add_ps_16xN_neon 32 4
+
+.macro pixel_add_ps_32xN_neon h i
+ function x265_pixel_add_ps_32x\h\()_neon
+    lsl             x5, x5, #1
+    mov             w12, #\i
+.loop_add_ps_32x\h\():
+    sub             w12, w12, #1
+.rept 4
+    ld1             {v0.16b-v1.16b}, [x2], x4
+    ld1             {v16.8h-v19.8h}, [x3], x5
+    uxtl            v4.8h, v0.8b
+    uxtl2           v5.8h, v0.16b
+    uxtl            v6.8h, v1.8b
+    uxtl2           v7.8h, v1.16b
+    add             v24.8h, v4.8h, v16.8h
+    add             v25.8h, v5.8h, v17.8h
+    add             v26.8h, v6.8h, v18.8h
+    add             v27.8h, v7.8h, v19.8h
+    sqxtun          v4.8b, v24.8h
+    sqxtun2         v4.16b, v25.8h
+    sqxtun          v5.8b, v26.8h
+    sqxtun2         v5.16b, v27.8h
+    st1             {v4.16b-v5.16b}, [x0], x1
+.endr
+    cbnz            w12, .loop_add_ps_32x\h
+    ret
+endfunc
+.endm
+
+pixel_add_ps_32xN_neon 32 8
+pixel_add_ps_32xN_neon 64 16
+
+function x265_pixel_add_ps_64x64_neon
+    lsl             x5, x5, #1
+    sub             x5, x5, #64
+    mov             w12, #32
+.loop_add_ps_64x64:
+    sub             w12, w12, #1
+.rept 2
+    ld1             {v0.16b-v3.16b}, [x2], x4
+    ld1             {v16.8h-v19.8h}, [x3], #64
+    ld1             {v20.8h-v23.8h}, [x3], x5
+    uxtl            v4.8h, v0.8b
+    uxtl2           v5.8h, v0.16b
+    uxtl            v6.8h, v1.8b
+    uxtl2           v7.8h, v1.16b
+    uxtl            v24.8h, v2.8b
+    uxtl2           v25.8h, v2.16b
+    uxtl            v26.8h, v3.8b
+    uxtl2           v27.8h, v3.16b
+    add             v0.8h, v4.8h, v16.8h
+    add             v1.8h, v5.8h, v17.8h
+    add             v2.8h, v6.8h, v18.8h
+    add             v3.8h, v7.8h, v19.8h
+    add             v4.8h, v24.8h, v20.8h
+    add             v5.8h, v25.8h, v21.8h
+    add             v6.8h, v26.8h, v22.8h
+    add             v7.8h, v27.8h, v23.8h
+    sqxtun          v0.8b, v0.8h
+    sqxtun2         v0.16b, v1.8h
+    sqxtun          v1.8b, v2.8h
+    sqxtun2         v1.16b, v3.8h
+    sqxtun          v2.8b, v4.8h
+    sqxtun2         v2.16b, v5.8h
+    sqxtun          v3.8b, v6.8h
+    sqxtun2         v3.16b, v7.8h
+    st1             {v0.16b-v3.16b}, [x0], x1
+.endr
+    cbnz            w12, .loop_add_ps_64x64
+    ret
+endfunc
+
+// Chroma add_ps
+function x265_pixel_add_ps_4x8_neon
+    lsl             x5, x5, #1
+.rept 4
+    ld1             {v0.8b}, [x2], x4
+    ld1             {v1.8b}, [x2], x4
+    ld1             {v2.4h}, [x3], x5
+    ld1             {v3.4h}, [x3], x5
+    uxtl            v0.8h, v0.8b
+    uxtl            v1.8h, v1.8b
+    add             v4.8h, v0.8h, v2.8h
+    add             v5.8h, v1.8h, v3.8h
+    sqxtun          v4.8b, v4.8h
+    sqxtun          v5.8b, v5.8h
+    st1             {v4.s}[0], [x0], x1
+    st1             {v5.s}[0], [x0], x1
+.endr
+    ret
+endfunc
+
+function x265_pixel_add_ps_8x16_neon
+    lsl             x5, x5, #1
+.rept 8
+    ld1             {v0.8b}, [x2], x4
+    ld1             {v1.8b}, [x2], x4
+    ld1             {v2.8h}, [x3], x5
+    ld1             {v3.8h}, [x3], x5
+    uxtl            v0.8h, v0.8b
+    uxtl            v1.8h, v1.8b
+    add             v4.8h, v0.8h, v2.8h
+    add             v5.8h, v1.8h, v3.8h
+    sqxtun          v4.8b, v4.8h
+    sqxtun          v5.8b, v5.8h
+    st1             {v4.8b}, [x0], x1
+    st1             {v5.8b}, [x0], x1
+.endr
+    ret
+endfunc
+
 .macro x265_satd_4x8_8x4_end_neon
     add             v0.8h, v4.8h, v6.8h
     add             v1.8h, v5.8h, v7.8h
-- 
2.25.1

